@@ -1,26 +1,25 @@
-try:
-    import requests
-    import re
-    import os
-    import json
-    import time
-    import logging
-    import sys
-    import json
-    import os
-    import websocket
-except ModuleNotFoundError as e:
-    os.chdir(os.getcwd())
-    os.system('pip install -r requirements.txt')
-try:
-    ws = websocket.WebSocket()
-except TypeError:
-    os.system('pip install --force-reinstall websocket-client')
-logging.basicConfig(filename='s2py.log', level=logging.INFO)
+'''
+scratch2py
+~~~~~~~~~~
+Scratch2py or S2py is an easy-to-use, versatile tool to communicate with the Scratch API and do other things
+   >>> from scratch2py import Scratch2Py
+   >>> s2py = Scratch2Py('foobar', 'spameggs')
+   >>> user = s2py.user('Scratchteam')
+   >>> user.exists()
+   True
+'''
 
+import websocket
+import requests
+import logging
+import json
+import re
+
+ws = websocket.WebSocket()
 
 class Scratch2Py():
-    def __init__(self, username, password):
+    '''Initialize a new Scratch2Py object'''
+    def __init__(self, username: str, password: str) -> None:
         self.chars = """AabBCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz0123456789 -_`~!@#$%^&*()+=[];:'"\|,.<>/?}{"""
         global uname
         uname = username
@@ -57,7 +56,7 @@ class Scratch2Py():
             ).group(1)
 
         except AttributeError:
-            sys.exit('Error: Invalid credentials. Authentication failed.')
+            raise Exception('Error: Invalid credentials. Authentication failed.')
         else:
             self.headers = {
                 "x-csrftoken": self.csrftoken,
@@ -72,6 +71,7 @@ class Scratch2Py():
             }
 
     def decode(self, text):
+        '''Encode a value.'''
         decoded = ""
         text = str(text)
         y = 0
@@ -82,9 +82,10 @@ class Scratch2Py():
         return decoded
 
     def encode(self, text):
+        '''Decode a value.'''
         encoded = ""
         length = int(len(text))
-        for i in range(0,length):
+        for i in range(0, length):
             try:
                 x = int(self.chars.index(text[i])+1)
                 if x < 10:
@@ -93,37 +94,21 @@ class Scratch2Py():
             except ValueError:
                 logging.error('Character not supported')
         return encoded
+
     class project:
-        def __init__(self, id):
+        '''Get project related information'''
+        def __init__(self, id: int) -> None:
             self.id = id
 
-        def getStats(self, stat):
-            if stat == "loves" or stat == "faves" or stat == "views" or stat == "remixes":
-                if stat == "loves":
-                    r = requests.get(
-                        "https://api.scratch.mit.edu/projects/"+str(self.id))
-                    data = r.json()
-                    return data['stats']['loves']
-                else:
-                    if stat == "faves":
-                        r = requests.get(
-                            "https://api.scratch.mit.edu/projects/"+str(self.id))
-                        data = r.json()
-                        return data['stats']['favorites']
-                    else:
-                        if stat == "remixes":
-                            r = requests.get(
-                                "https://api.scratch.mit.edu/projects/"+str(self.id))
-                            data = r.json()
-                            return data['stats']['remixes']
-                        else:
-                            if stat == "views":
-                                r = requests.get(
-                                    "https://api.scratch.mit.edu/projects/"+str(self.id))
-                                data = r.json()
-                                return data['stats']['views']
+        def getStats(self) -> dict:
+            '''Fetch statistics for a project'''
+            r = requests.get(
+                "https://api.scratch.mit.edu/projects/"+str(self.id))
+            data = r.json()
+            return data['stats']
 
-        def getComments(self):
+        def getComments(self) -> 'list[dict]':
+            '''Fetch comments for a project'''
             uname = requests.get(
                 "https://api.scratch.mit.edu/projects/"+str(self.id)).json()
             if uname != {"code": "NotFound", "message": ""}:
@@ -145,14 +130,16 @@ class Scratch2Py():
                             str('Username: '+str(uname))+(str(', Content: ')+str(x)))
                     return data
 
-        def getInfo(self):
+        def getInfo(self) -> dict:
+            '''Fetch the entire project information'''
             r = requests.get(
                 'https://api.scratch.mit.edu/projects/'+str(self.id)
             ).json()
             return r
 
-        def fetchAssets(self, type='img'):
+        def getAssets(self, type='img') -> 'list[str]':
             '''
+            Fetch project asset URLs such as their costumes and sounds
             You may have problems with fetching assets since some projects may not have any assets, or are fetched as binary code and not JSON
             '''
 
@@ -171,7 +158,8 @@ class Scratch2Py():
             return assets
 
     class studioSession:
-        def __init__(self, sid):
+        '''Get studio related information'''
+        def __init__(self, sid: int) -> None:
             self.headers = {
                 "x-csrftoken": "a",
                 "x-requested-with": "XMLHttpRequest",
@@ -181,18 +169,21 @@ class Scratch2Py():
             }
             self.sid = sid
 
-        def inviteCurator(self, user):
+        def inviteCurator(self, user: str) -> None:
+            '''Invite a curator to a studio'''
             self.headers["referer"] = (
                 "https://scratch.mit.edu/studios/" + str(self.sid) + "/curators/")
             requests.put("https://scratch.mit.edu/site-api/users/curators-in/" +
                          str(self.sid) + "/invite_curator/?usernames=" + user, headers=self.headers)
 
-        def addStudioProject(self, pid):
+        def addStudioProject(self, pid: int):
+            '''Add a project to a studio'''
             self.headers['referer'] = "https://scratch.mit.edu/projects/" + \
                 str(pid)
             return requests.post("https://api.scratch.mit.edu/studios/"+str(self.sid)+"/project/"+str(pid), headers=self.headers)
 
-        def postComment(self, content, parent_id="", commentee_id=""):
+        def postComment(self, content: str, parent_id: str="", commentee_id: str=""):
+            '''Post a comment on a studio'''
             self.headers['referer'] = (
                 "https://scratch.mit.edu/studios/" +
                 str(self.sid) + "/comments/"
@@ -210,7 +201,8 @@ class Scratch2Py():
                 data=json.dumps(data),
             )
 
-        def getComments(self):
+        def getComments(self) -> 'list[str]':
+            '''Fetch studio comments'''
             r = requests.get(
                 "https://api.scratch.mit.edu/studios/"+str(self.sid)+"/comments")
             data = r.json()
@@ -218,9 +210,10 @@ class Scratch2Py():
             for i in data:
                 x = i['content']
                 comments.append(x)
-            return json.dumps(comments)
+            return comments
 
-        def follow(self):
+        def follow(self) -> dict:
+            '''Follow a studio'''
             self.headers['referer'] = "https://scratch.mit.edu/studios/" + \
                 str(self.sid)
             return requests.put(
@@ -230,8 +223,9 @@ class Scratch2Py():
                 + self.username,
                 headers=self.headers,
             ).json()
-
-        def unfollow(self):
+    
+        def unfollow(self) -> dict:
+            '''Unfollow a studio'''
             self.headers['referer'] = "https://scratch.mit.edu/studios/" + \
                 str(self.sid)
             return requests.put(
@@ -243,7 +237,8 @@ class Scratch2Py():
             ).json()
 
     class projectSession:
-        def __init__(self, pid):
+        '''Do project related things as a "user"'''
+        def __init__(self, pid: int) -> None:
             self.headers = {
                 "x-csrftoken": "a",
                 "x-requested-with": "XMLHttpRequest",
@@ -254,18 +249,21 @@ class Scratch2Py():
             self.pid = pid
 
         def share(self):
+            '''Share a project'''
             self.headers["referer"] = (
                 "https://scratch.mit.edu/projects/"+str(self.pid)
             )
             return requests.put("https://api.scratch.mit.edu/proxy/projects/"+str(self.pid)+"/share", headers=self.headers)
 
         def unshare(self):
+            '''Unshare a project'''
             self.headers["referer"] = (
                 "https://scratch.mit.edu/projects/"+str(self.pid)
             )
             return requests.put("https://api.scratch.mit.edu/proxy/projects/"+str(self.pid)+"/unshare", headers=self.headers)
 
         def favorite(self):
+            '''Favorite (star) a project'''
             self.headers['referer'] = "https://scratch.mit.edu/projects/" + \
                 str(self.pid)
             return requests.post(
@@ -277,6 +275,7 @@ class Scratch2Py():
             ).json()
 
         def unfavorite(self):
+            '''Unfavorite (star) a project'''
             self.headers['referer'] = "https://scratch.mit.edu/projects/" + \
                 str(self.pid)
             return requests.delete(
@@ -288,6 +287,7 @@ class Scratch2Py():
             ).json()
 
         def love(self):
+            '''Love (heart) a project'''
             self.headers['referer'] = "https://scratch.mit.edu/projects/" + \
                 str(self.pid)
             return requests.post(
@@ -299,6 +299,7 @@ class Scratch2Py():
             ).json()
 
         def unlove(self):
+            '''Unlove (heart) a project'''
             self.headers['referer'] = "https://scratch.mit.edu/projects/" + \
                 str(self.pid)
             return requests.delete(
@@ -309,8 +310,15 @@ class Scratch2Py():
                 headers=self.headers,
             ).json()
 
+        def remix(self):
+            '''Remix a project'''
+            self.headers['referer'] = "https://scratch.mit.edu/projects/" + \
+                str(self.pid)
+            return requests.post("https://projects.scratch.mit.edu/?is_remix=1&original_id="+str(self.pid)+"&title=Scratch%20Project").json()
+
     class userSession:
-        def __init__(self, username):
+        '''Do profile related things as a "user"'''
+        def __init__(self, username: str) -> None:
             self.headers = {
                 "x-csrftoken": "a",
                 "x-requested-with": "XMLHttpRequest",
@@ -322,6 +330,7 @@ class Scratch2Py():
             self.uname2 = username
 
         def followUser(self):
+            '''Follow a user'''
             self.headers['referer'] = "https://scratch.mit.edu/users/" + \
                 str(self.username)+"/"
             return requests.put(
@@ -333,6 +342,7 @@ class Scratch2Py():
             ).json()
 
         def unfollowUser(self):
+            '''Unfollow a user'''
             self.headers['referer'] = "https://scratch.mit.edu/users/" + \
                 str(self.username)+"/"
             return requests.put(
@@ -344,6 +354,7 @@ class Scratch2Py():
             ).json()
 
         def toggleCommenting(self):
+            '''Toggle a user profile's comment section'''
             self.headers['referer'] = "https://scratch.mit.edu/users/" + \
                 str(self.username)
             return requests.post(
@@ -352,17 +363,19 @@ class Scratch2Py():
                 headers=self.headers,
             )
 
-        def postComment(self,content, parent_id="", commentee_id=""):
+        def postComment(self, content: str, parent_id: str="", commentee_id: str=""):
+            '''Post a comment on a user profile'''
             self.headers['referer'] = "https://scratch.mit.edu/users/" + self.uname2
             data = {
                 'content': content,
                 'parent_id': parent_id,
                 'commentee_id': commentee_id
             }
-            return requests.post("https://scratch.mit.edu/site-api/comments/user/"+ self.uname2 +"/add/",data=json.dumps(data),headers=self.headers).json()
+            return requests.post("https://scratch.mit.edu/site-api/comments/user/" + self.uname2 + "/add/", data=json.dumps(data), headers=self.headers).json()
 
     class user:
-        def __init__(self, user):
+        '''Do profile related things'''
+        def __init__(self, user: str) -> None:
             self.user = user
             self.headers = {
                 "x-csrftoken": "a",
@@ -372,36 +385,42 @@ class Scratch2Py():
                 "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36"
             }
 
-        def exists(self):
+        def exists(self) -> bool:
+            '''Check if a user exists. Returns true or false'''
             return requests.get("https://api.scratch.mit.edu/accounts/checkusername/"+str(self.user)).json() == {"username": self.user, "msg": "username exists"}
 
         def getMessagesCount(self):
+            '''Get the number of messages someone posted'''
             self.headers['referer'] = "https://scratch.mit.edu"
             return requests.get("https://api.scratch.mit.edu/users/"+str(self.user)+"/messages/count").json()['count']
 
-        def getMessages(self):
+        def getMessages(self) -> dict:
+            '''Get the messages someone sent'''
             return requests.get("https://api.scratch.mit.edu/users/"+str(self.user)+"/messages" + "/", headers=self.headers).json()
 
-        def getStatus(self):
+        def getStatus(self) -> str:
+            '''Get the "About Me" section of a user profile'''
             return requests.get("https://api.scratch.mit.edu/users/"+str(self.user)).json()['profile']['status']
 
-        def getBio(self):
+        def getBio(self) -> str:
+            '''Get the 'What I'm Working On' section of a user profile'''
             return requests.get("https://api.scratch.mit.edu/users/"+str(self.user)).json()['profile']['bio']
 
-        def getProjects(self):
+        def getProjects(self) -> 'dict[str, list]':
             r = requests.get(
                 "https://api.scratch.mit.edu/users/"+str(self.user)+"/projects")
             data = r.json()
-            titles = []
+            titles = {}
             for i in data:
                 x = i['title']
                 y = i['id']
-                titles.append('ID: ' + str(y))
-                titles.append('Title: ' + str(x))
+                z = i['history']['shared']
+                titles[x] = [y, z]
             return titles
 
     class scratchConnect:
-        def __init__(self, pid):
+        '''Connect to a project to read and write cloud variables'''
+        def __init__(self, pid: int):
             global ws
             global PROJECT_ID
             self.username = uname
@@ -414,7 +433,8 @@ class Scratch2Py():
                 'project_id': str(pid)
             }) + '\n')
 
-        def setCloudVar(self, variable, value):
+        def setCloudVar(self, variable: str, value: str) -> None:
+            '''Set the value of a cloud variable (☁ is not needed)'''
             try:
                 ws.send(json.dumps({
                     'method': 'set',
@@ -442,7 +462,8 @@ class Scratch2Py():
                     'project_id': str(PROJECT_ID)
                 }) + '\n')
 
-        def readCloudVar(self, name, limit="1000"):
+        def readCloudVar(self, name: str, limit: str="1000") -> str:
+            '''Read the value of a cloud variable (an exception may be raised if reading fails)'''
             try:
                 resp = requests.get("https://clouddata.scratch.mit.edu/logs?projectid=" +
                                     str(PROJECT_ID)+"&limit="+str(limit)+"&offset=0").json()
@@ -450,11 +471,12 @@ class Scratch2Py():
                     x = i['name']
                     if x == ('☁ ' + str(name)):
                         return i['value']
-            except:
-                return 'Sorry, there was an error.'
+            except Exception:
+                raise Exception('Cloud variable could not be read.')
 
     class scratchDatabase:
-        def __init__(self, pid):
+        '''Create a new database that will detect messages on a certain project ID'''
+        def __init__(self, pid: int) -> None:
             self.chars = """AabBCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz0123456789 -_`~!@#$%^&*()+=[];:'"\|,.<>/?}{"""
             self.id = pid
             self.username = uname
@@ -479,7 +501,7 @@ class Scratch2Py():
         def __encode(self, text):
             encoded = ""
             length = int(len(text))
-            for i in range(0,length):
+            for i in range(0, length):
                 try:
                     x = int(self.chars.index(text[i])+1)
                     if x < 10:
@@ -534,6 +556,7 @@ class Scratch2Py():
                         return i['value']
 
         def startLoop(self):
+            '''Start a new loop for the database'''
             data = []
             while True:
                 encodedMethod = self.__readCloudVar('Method')
@@ -606,192 +629,4 @@ class Scratch2Py():
                         self.__setCloudVar('Return', tosend)
                     else:
                         tosend = self.__encode('Invalid input.')
-                        self.__setCloudVar('Return', tosend)
-
-    class turbowarpDatabase:
-        def __init__(self, pid):
-            self.chars = """AabBCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz0123456789 -_`~!@#$%^&*()+=[];:'"\|,.<>/?}{"""
-            self.id = pid
-            self.username = uname
-            ws.connect('wss://clouddata.turbowarp.org',
-                       origin='https://turbowarp.org', enable_multithread=True)
-            ws.send(json.dumps({
-                'method': 'handshake',
-                'user': self.username,
-                'project_id': str(self.id)
-            }) + '\n')
-
-        def __decode(self, text):
-            decoded = ""
-            text = str(text)
-            y = 0
-            for i in range(0, len(text)//2):
-                x = self.chars[int(str(text[y])+str(text[int(y)+1]))-1]
-                decoded = str(decoded)+str(x)
-                y += 2
-            return decoded
-
-        def __encode(self, text):
-            encoded = ""
-            length = int(len(text))
-            for i in range(0,length):
-                try:
-                    x = int(self.chars.index(text[i])+1)
-                    if x < 10:
-                        x = str(0)+str(x)
-                    encoded = encoded + str(x)
-                except ValueError:
-                    logging.error('Character not supported')
-            return encoded
-
-        def __readCloudVar(self, variable):
-            ws.send(json.dumps({
-                'method': 'get',
-                'project_id': str(turbowarpid)
-            }) + '\n')
-            data = ws.recv()
-            data = data.split('\n')
-            result = []
-            for i in data:
-                result.append(json.loads(i))
-            for i in result:
-                if i['name'] == '☁ ' + variable:
-                    return i['value']
-
-        def __setCloudVar(self, variable, value):
-            ws.send(json.dumps({
-                'method': 'set',
-                'name': '☁ ' + variable,
-                'value': str(value),
-                'user': self.username,
-                'project_id': str(turbowarpid)
-            }) + '\n')
-
-        def __readCloudVar(self, name, limit="1000"):
-            try:
-                resp = requests.get("https://clouddata.scratch.mit.edu/logs?projectid=" +
-                                    str(self.id)+"&limit="+str(limit)+"&offset=0").json()
-                for i in resp:
-                    x = i['name']
-                    if x == ('☁ ' + str(name)):
-                        return i['value']
-            except json.decoder.JSONDecodeError:
-                resp = requests.get("https://clouddata.scratch.mit.edu/logs?projectid=" +
-                                    str(self.id)+"&limit="+str(limit)+"&offset=0").json()
-                for i in resp:
-                    x = i['name']
-                    if x == ('☁ ' + str(name)):
-                        return i['value']
-
-        def startLoop(self):
-            data = []
-            while True:
-                encodedMethod = self.__readCloudVar('Method')
-                if encodedMethod != None:
-                    Method = self.__decode(encodedMethod)
-                if Method == "set":
-                    encodedSend = self.__readCloudVar('Send')
-                    Send = str(self.__decode(encodedSend))
-                    encodedVal = self.__readCloudVar('Data')
-                    Val = str(self.__decode(encodedVal))
-                    intVal = self.__decode(encodedVal)
-                    c = 0
-                    for i in Send:
-                        if str(i) in "1234567890":
-                            c = int(c)+1
-                    if c == len(Send):
-                        if int(Send) > len(data):
-                            if int(Send) == int(len(data))+1:
-                                data.append(intVal)
-                                logging.info('Data added.')
-                                tosend = self.__encode('Data added.')
-                                self.__setCloudVar('Return', tosend)
-                                self.__setCloudVar('Method', '')
-                            else:
-                                while len(data) != int(Send)-1:
-                                    data.append('none')
-                                data.append(intVal)
-                                logging.info('Data added.')
-                                tosend = self.__encode('Data added.')
-                                self.__setCloudVar('Return', tosend)
-                                self.__setCloudVar('Method', '')
-                        else:
-                            data.pop(int(Send)-1)
-                            data.insert(int(Send), intVal)
-                            logging.info('Data added.')
-                            tosend = self.__encode('Data added.')
-                            self.__setCloudVar('Return', tosend)
-                            self.__setCloudVar('Method', '')
-                    else:
-                        tosend = self.__encode(
-                            'Invalid input. Variable name must be int.')
-                        self.__setCloudVar('Return', tosend)
-                if Method == "get":
-                    encodedSend = self.__readCloudVar('Send')
-                    Send = self.__decode(encodedSend)
-                    c = 0
-                    for i in Send:
-                        if str(i) in "1234567890":
-                            c = int(c)+1
-                    if c == len(Send) and int(Send) > 0 and int(Send) < int(len(data))+1:
-                        tosend = self.__encode(data[int(Send)-1])
-                        self.__setCloudVar('Return', tosend)
-                        logging.info('Data sent.')
-                        self.__setCloudVar('Method', '')
-                    else:
-                        tosend = self.__encode('Invalid input.')
-                        self.__setCloudVar('Return', tosend)
-                if Method == "delete":
-                    encodedSend = self.__readCloudVar('Send')
-                    Send = self.__decode(encodedSend)
-                    c = 0
-                    for i in Send:
-                        if str(i) in "1234567890":
-                            c = int(c)+1
-                    if c == len(Send) and int(Send) > 0 and int(Send) < int(len(data))+1:
-                        data.pop(int(Send)-1)
-                        data.insert(int(Send)-1, 'none')
-                        logging.info('Variable deleted.')
-                        tosend = self.__encode('Variable deleted.')
-                        self.__setCloudVar('Return', tosend)
-                    else:
-                        tosend = self.__encode('Invalid input.')
-                        self.__setCloudVar('Return', tosend)
-
-    class turbowarpConnect:
-        def __init__(self, pid):
-            global ws
-            global turbowarpid
-            self.username = uname
-            turbowarpid = pid
-            ws.connect('wss://clouddata.turbowarp.org',
-                       origin='https://turbowarp.org', enable_multithread=True)
-            ws.send(json.dumps({
-                'method': 'handshake',
-                'user': self.username,
-                'project_id': str(turbowarpid)
-            }) + '\n')
-
-        def setTurbowarpVar(self, variable, value):
-            ws.send(json.dumps({
-                'method': 'set',
-                'name': '☁ ' + variable,
-                'value': str(value),
-                'user': self.username,
-                'project_id': str(turbowarpid)
-            }) + '\n')
-
-        def readTurbowarpVar(self, variable):
-            ws.send(json.dumps({
-                'method': 'get',
-                'project_id': str(turbowarpid)
-            }) + '\n')
-            data = ws.recv()
-            data = data.split('\n')
-            result = []
-            for i in data:
-                result.append(json.loads(i))
-            for i in result:
-                if i['name'] == '☁ ' + variable:
-                    return i['value']
-            return 'Variable not found.'
+                        self.__setCloudVar('Return', tosend)
\ No newline at end of file
